/* -*- Mode: C++; c-default-style: "k&r"; indent-tabs-mode: nil; tab-width: 2; c-basic-offset: 2 -*- */

/* libmwaw
* Version: MPL 2.0 / LGPLv2+
*
* The contents of this file are subject to the Mozilla Public License Version
* 2.0 (the "License"); you may not use this file except in compliance with
* the License or as specified alternatively below. You may obtain a copy of
* the License at http://www.mozilla.org/MPL/
*
* Software distributed under the License is distributed on an "AS IS" basis,
* WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
* for the specific language governing rights and limitations under the
* License.
*
* Major Contributor(s):
* Copyright (C) 2002 William Lachance (wrlach@gmail.com)
* Copyright (C) 2002,2004 Marc Maurer (uwog@uwog.net)
* Copyright (C) 2004-2006 Fridrich Strba (fridrich.strba@bluewin.ch)
* Copyright (C) 2006, 2007 Andrew Ziem
* Copyright (C) 2011, 2012 Alonso Laurent (alonso@loria.fr)
*
*
* All Rights Reserved.
*
* For minor contributions see the git repository.
*
* Alternatively, the contents of this file may be used under the terms of
* the GNU Lesser General Public License Version 2 or later (the "LGPLv2+"),
* in which case the provisions of the LGPLv2+ are applicable
* instead of those above.
*/

#include <cmath>
#include <iomanip>
#include <iostream>
#include <limits>
#include <sstream>
#include <utility>

#include <librevenge/librevenge.h>

#include "MWAWGraphicListener.hxx"
#include "MWAWGraphicShape.hxx"
#include "MWAWGraphicStyle.hxx"
#include "MWAWHeader.hxx"
#include "MWAWParagraph.hxx"
#include "MWAWPictBitmap.hxx"
#include "MWAWPictData.hxx"
#include "MWAWPrinter.hxx"
#include "MWAWPosition.hxx"
#include "MWAWSubDocument.hxx"

#include "SpringBoardParser.hxx"

/** Internal: the structures of a SpringBoardParser */
namespace SpringBoardParserInternal
{
//! Internal: a structure used to store the frame property in SpringBoardParserInternal
struct Frame {
  //! constructor
  Frame()
    : m_type(-1)
    , m_page(1)
    , m_pageId(0)
    , m_style(MWAWGraphicStyle::emptyStyle())
    , m_numSubZones(0)
  {
  }
  //! the frame type: 0: layout, 1: text, 2: graphic
  int m_type;
  //! the page number
  int m_page;
  //! the page frame id
  int m_pageId;
  //! the frame style
  MWAWGraphicStyle m_style;
  //! the number of sub zone (for a text frame)
  int m_numSubZones;

};

////////////////////////////////////////
//! Internal: the state of a SpringBoardParser
struct State {
  //! constructor
  State()
    : m_numPages(1)
    , m_entriesListMap()
    , m_entries2ListMap()
    , m_idToFrameMap()
  {
  }

  //! returns the name of a data zone
  std::string getDataZoneName(int id) const
  {
    char const *wh[]= {"Doc", "Page", "FramStyle", "FramLink", nullptr,
                       nullptr, nullptr, nullptr, "ParaStyle", "Text"
                      };
    if (id>=0 && id<10 && wh[id])
      return wh[id];
    std::stringstream s;
    s << "List" << id;
    return s.str();
  }
  //! the number of pages
  int m_numPages;
  //! a map zone id to list's entry
  std::map<int, MWAWEntry> m_entriesListMap;
  //! a map id to the last zone entries
  std::map<int, MWAWEntry> m_entries2ListMap;
  //! the map id to frame map
  std::map<int, Frame> m_idToFrameMap;
};

////////////////////////////////////////
//! Internal: the subdocument of a SpringBoardParser
class SubDocument final : public MWAWSubDocument
{
public:
  SubDocument(SpringBoardParser &pars, MWAWInputStreamPtr const &input)
    : MWAWSubDocument(&pars, input, MWAWEntry())
  {
  }

  //! destructor
  ~SubDocument() final {}

  //! operator!=
  bool operator!=(MWAWSubDocument const &doc) const final
  {
    if (MWAWSubDocument::operator!=(doc)) return true;
    auto const *sDoc = dynamic_cast<SubDocument const *>(&doc);
    if (!sDoc) return true;
    return false;
  }

  //! the parser function
  void parse(MWAWListenerPtr &listener, libmwaw::SubDocumentType type) final;

protected:
private:
  SubDocument(SubDocument const &orig) = delete;
  SubDocument &operator=(SubDocument const &orig) = delete;
};

void SubDocument::parse(MWAWListenerPtr &listener, libmwaw::SubDocumentType)
{
  if (!listener || !listener->canWriteText()) {
    MWAW_DEBUG_MSG(("SpringBoardParserInternal::SubDocument::parse: no listener\n"));
    return;
  }
  if (!m_parser) {
    MWAW_DEBUG_MSG(("SpringBoardParserInternal::SubDocument::parse: no parser\n"));
    return;
  }
  long pos = m_input->tell();
  //static_cast<SpringBoardParser *>(m_parser)->sendText(m_id, m_subId);
  m_input->seek(pos, librevenge::RVNG_SEEK_SET);
}


}

////////////////////////////////////////////////////////////
// constructor/destructor, ...
////////////////////////////////////////////////////////////
SpringBoardParser::SpringBoardParser(MWAWInputStreamPtr const &input, MWAWRSRCParserPtr const &rsrcParser, MWAWHeader *header)
  : MWAWGraphicParser(input, rsrcParser, header)
  , m_state(new SpringBoardParserInternal::State)
{
  setAsciiName("main-1");

  getPageSpan().setMargins(0.1);
}

SpringBoardParser::~SpringBoardParser()
{
}

////////////////////////////////////////////////////////////
// the parser
////////////////////////////////////////////////////////////
void SpringBoardParser::parse(librevenge::RVNGDrawingInterface *docInterface)
{
  if (!getInput().get() || !checkHeader(nullptr))  throw(libmwaw::ParseException());
  bool ok = false;
  try {
    // create the asciiFile
    ascii().setStream(getInput());
    ascii().open(asciiName());
    checkHeader(nullptr);
    ok = createZones();
    if (ok) {
      createDocument(docInterface);

      auto listener=getGraphicListener();
    }
    ascii().reset();
  }
  catch (...) {
    MWAW_DEBUG_MSG(("SpringBoardParser::parse: exception catched when parsing\n"));
    ok = false;
  }

  resetGraphicListener();
  if (!ok) throw(libmwaw::ParseException());
}

////////////////////////////////////////////////////////////
// create the document
////////////////////////////////////////////////////////////
void SpringBoardParser::createDocument(librevenge::RVNGDrawingInterface *documentInterface)
{
  if (!documentInterface) return;
  if (getGraphicListener()) {
    MWAW_DEBUG_MSG(("SpringBoardParser::createDocument: listener already exist\n"));
    return;
  }

  // create the page list
  std::vector<MWAWPageSpan> pageList;
  int numPages=std::max<int>(1,m_state->m_numPages);
  MWAWPageSpan ps(getPageSpan());
  ps.setPageSpan(numPages);
  pageList.push_back(ps);

  MWAWGraphicListenerPtr listen(new MWAWGraphicListener(*getParserState(), pageList, documentInterface));
  setGraphicListener(listen);
  listen->startDocument();
}


////////////////////////////////////////////////////////////
//
// Intermediate level
//
////////////////////////////////////////////////////////////
bool SpringBoardParser::createZones()
{
  MWAWInputStreamPtr input = getInput();
  input->seek(4, librevenge::RVNG_SEEK_SET);

  libmwaw::DebugStream f;
  f << "limits=[";
  for (int i=0; i<2; ++i) {
    long ptr=long(input->readULong(4));
    f << std::hex << ptr << std::dec << ",";
    if (ptr<0x80 || !input->checkPosition(ptr)) {
      MWAW_DEBUG_MSG(("SpringBoardParser::createZones: find bad limits\n"));
      f << "###";
      ascii().addPos(0);
      ascii().addNote(f.str().c_str());
      return false;
    }
    ascii().addPos(ptr);
    ascii().addNote(i==0 ? "[A]" : "[B]");
  }
  f << "],";

  int val;
  MWAWEntry entry;
  entry.setBegin(long(input->readULong(4)));
  entry.setLength(long(input->readULong(4)));
  if (entry.valid()) {
    f << "pagNumb=" << std::hex << entry.begin() << "<->" << entry.end() << std::dec << ",";
    if (entry.begin()<0x80 || !input->checkPosition(entry.end())) {
      MWAW_DEBUG_MSG(("SpringBoardParser::createZones: find bad entries for page numbering\n"));
      f << "###";
    }
    else {
      long actPos=input->tell();
      parsePageNumbering(entry);
      input->seek(actPos, librevenge::RVNG_SEEK_SET);
    }
  }
  for (int i=0; i<10; ++i) {
    entry.setBegin(long(input->readULong(4)));
    entry.setLength(long(input->readULong(2)));
    if (!entry.valid())
      continue;
    char const *wh[]= {
      "doc", "page", "framstyle", "framlink", nullptr,
      nullptr, nullptr, nullptr, "parastyle", "text"
    };
    if (wh[i])
      f << wh[i] << "=" << std::hex << entry.begin() << "<->" << entry.end() << std::dec << ",";
    else
      f << "list" << i << "=" << std::hex << entry.begin() << "<->" << entry.end() << std::dec << ",";
    if (entry.begin()<0x80 || !input->checkPosition(entry.end())) {
      MWAW_DEBUG_MSG(("SpringBoardParser::createZones: find bad entries %d\n", i));
      f << "###";
      ascii().addPos(0);
      ascii().addNote(f.str().c_str());
      return false;
    }
    m_state->m_entriesListMap[i]=entry;
  }
  for (int i=0; i<3; ++i) {
    val=int(input->readLong(2));
    if (val)
      f << "f" << i+4 << "=" << val << ",";
  }
  for (int i=0; i<3; ++i) {
    entry.setBegin(long(input->readULong(4)));
    entry.setLength(long(input->readULong(4)));
    if (!entry.valid())
      continue;
    f << (i==0 ? "shape" : i==1 ? "txtPos" : "frame") << "=" << std::hex << entry.begin() << "<->" << entry.end() << std::dec << ",";
    if (entry.begin()<0x80 || !input->checkPosition(entry.end())) {
      MWAW_DEBUG_MSG(("SpringBoardParser::createZones: find bad entries %d\n", i));
      f << "###";
      ascii().addPos(0);
      ascii().addNote(f.str().c_str());
      return false;
    }
    m_state->m_entries2ListMap[i]=entry;
  }
  for (int i=0; i<3; ++i) {
    val=int(input->readLong(2));
    if (val)
      f << "g" << i << "=" << val << ",";
  }
  val=int(input->readULong(4)); // 0 or 68ad08
  if (val)
    f << "unkn=" << std::hex << val << std::dec << ",";
  for (int i=0; i<3; ++i) { // g3=0|8
    val=int(input->readLong(2));
    if (val)
      f << "g" << i+3 << "=" << val << ",";
  }
  f << "num[shapes]=" << input->readLong(2) << ",";
  ascii().addPos(0);
  ascii().addNote(f.str().c_str());

  for (auto const &it : m_state->m_entriesListMap)
    parseDataList(it.first, it.second);

  for (auto const &it : m_state->m_entries2ListMap)
    parseLastZone(it.first, it.second);
  return false;
}

bool SpringBoardParser::parsePageNumbering(MWAWEntry const &entry)
{
  MWAWInputStreamPtr input = getInput();
  if (!input)
    return false;
  if (!entry.valid() || (entry.length()%98)!=0 || !input->checkPosition(entry.end())) {
    MWAW_DEBUG_MSG(("SpringBoardParser::parsePageNumbering: find unvalid entry\n"));
    return false;
  }
  ascii().addPos(entry.end());
  ascii().addNote("_");

  libmwaw::DebugStream f;
  f << "Entries(PagNumbering):";
  input->seek(entry.begin(), librevenge::RVNG_SEEK_SET);
  ascii().addPos(entry.begin());
  ascii().addNote(f.str().c_str());
  int N=int(entry.length()/98);
  int val;
  for (int i=0; i<N; ++i) {
    long pos=input->tell();
    f.str("");
    f << "PagNumbering-" << i << ":";
    int pg=int(input->readLong(2));
    f << "page=" << pg << ",";
    for (int j=0; j<2; ++j) {
      val=int(input->readLong(2));
      if (val==2-j)
        continue;
      if (j==0) // 0: numeric, 1/2: small/big roman, 2/3: small/big alpha, 4-6: full minu,MAJU,Title, 8: none
        f << "format=" << val << ",";
      else
        f << "page[number]=" << val << ",";
    }
    val=int(input->readLong(2)); // TC, TR, TL, BC, BR, BL, FacingT, FacingB
    f << "pos=" << val << ",";
    f << "fId=" << int(input->readLong(2)) << ",";
    val=int(input->readULong(2));
    if (val)
      f << "font[fl]=" << std::hex << val << std::dec << ",";
    f << "fSize=" << int(input->readLong(2)) << ",";
    for (int j=0; j<2; ++j) {
      long actPos=input->tell();
      int cLen=int(input->readULong(1));
      if (cLen>40) {
        MWAW_DEBUG_MSG(("SpringBoardParser::parsePageNumbering: can not read a string\n"));
        f << "###cLen" << j << "=" << cLen << ",";
        cLen=0;
      }
      std::string text;
      for (int c=0; c<cLen; ++c) text+=char(input->readULong(1));
      f << text << ",";
      input->seek(actPos+40, librevenge::RVNG_SEEK_SET);
    }
    for (int j=0; j<2; ++j) {
      val=int(input->readLong(2));
      if (val!=(j==0 ? 50 : 90))
        f << "from[" << (j==0 ? "bottom" : "right") << "]=" << val << ",";
    }
    ascii().addPos(pos);
    ascii().addNote(f.str().c_str());
    input->seek(pos+98, librevenge::RVNG_SEEK_SET);
  }
  return true;
}

bool SpringBoardParser::parseDataList(int id, MWAWEntry const &entry)
{
  MWAWInputStreamPtr input = getInput();
  if (!input)
    return false;
  if (id<0 || id>=10 || !entry.valid() || !input->checkPosition(entry.end())) {
    MWAW_DEBUG_MSG(("SpringBoardParser::parseDataList: find unvalid entry for id=%d\n", id));
    return false;
  }
  ascii().addPos(entry.end());
  ascii().addNote("_");

  libmwaw::DebugStream f;
  std::string const what=m_state->getDataZoneName(id);
  f << "Entries(" << what << "):";
  if ((entry.length()%16)!=0) {
    f << "###";
    MWAW_DEBUG_MSG(("SpringBoardParser::parseDataList: find unexpected size for zone %s\n", what.c_str()));
    ascii().addPos(entry.begin());
    ascii().addNote(f.str().c_str());
    return false;
  }

  input->seek(entry.begin(), librevenge::RVNG_SEEK_SET);
  int num=int(entry.length()/16);
  std::map<int, std::pair<MWAWVec2i,MWAWEntry> > listZones;
  for (int st=0; st<num; ++st) {
    long pos=input->tell();
    f.str("");
    if (st==0)
      f << "Entries(" << what << "):";
    else
      f << what << "[st=" << st << "]:";
    int firstId=int(input->readLong(4));
    if (firstId)
      f << "id[first]=" << firstId << ",";
    MWAWEntry data;
    data.setBegin(long(input->readULong(4)));
    data.setLength(long(input->readULong(4)));
    f << "zone=" << std::hex << data.begin() << "x" << data.end() << std::dec << ",";
    int nextId=int(input->readLong(4));
    f << "next[id]=" << nextId << ",";
    if (!data.valid() || data.begin()<0x80 || !input->checkPosition(data.end())) {
      f << "###";
      MWAW_DEBUG_MSG(("SpringBoardParser::parseDataList: find unexpected zone for zone %s\n", what.c_str()));
      return false;
    }
    else
      listZones[st]=std::make_pair(MWAWVec2i(firstId,nextId), data);
    ascii().addPos(pos);
    ascii().addNote(f.str().c_str());
  }
  for (auto const &it : listZones) {
    MWAWEntry data;
    MWAWVec2i indices;
    std::tie(indices,data)=it.second;
    if (id==8)
      parseParagraphStyle(data, indices);
    else
      parseData(id, data, indices);
  }
  return true;
}

bool SpringBoardParser::parseData(int id, MWAWEntry const &entry, MWAWVec2i const &indices)
{
  MWAWInputStreamPtr input = getInput();
  if (!input)
    return false;
  if (id<0 || id>=10 || !entry.valid() || !input->checkPosition(entry.end())) {
    MWAW_DEBUG_MSG(("SpringBoardParser::parseData: find unvalid entry for id=%d\n", id));
    return false;
  }

  input->seek(entry.begin(), librevenge::RVNG_SEEK_SET);
  long pos=input->tell();
  ascii().addPos(entry.end());
  ascii().addNote("_");

  libmwaw::DebugStream f;
  std::string const what=m_state->getDataZoneName(id);
  int const expectedSize[]= {6, 10, 26, 182, -1,
                             -1, -1, -1, -1, 512
                            };
  if (expectedSize[id]<=0) {
    f << what << "[header,index=" << indices << "]:";
    ascii().addPos(pos);
    ascii().addNote(f.str().c_str());
    return true;
  }
  if ((indices[1]-indices[0])*expectedSize[id]>entry.length()) {
    MWAW_DEBUG_MSG(("SpringBoardParser::parseData: find unexpected length for data zone %s\n", what.c_str()));
    ascii().addPos(pos);
    ascii().addNote("###");
    return false;
  }
  for (int i=indices[0]; i<indices[1]; ++i) {
    pos=input->tell();
    f.str("");
    f << what << "-dt" << i << "]:";
    int val;
    switch (id) {
    case 0: { // Doc
      for (int j=0; j<3; ++j) {
        val=int(input->readLong(2));
        if (val==1)
          continue;
        if (j==2)
          f << "num[pages]=" << val << ",";
        else
          f << "num" << j << "=" << val << ",";
      }
      break;
    }
    case 1: { // Page
      for (int j=0; j<5; ++j) {
        val=int(input->readLong(2));
        int const expected[]= {0,0,0,0,-1};
        if (val==expected[j])
          continue;
        if (j==0)
          f << "page=" << val << ",";
        else if (j==3)
          f << "first[shape]=" << val << ",";
        else
          f << "f" << j << "=" << val << ",";
      }
      break;
    }
    case 2: { // FrameStyle
      SpringBoardParserInternal::Frame frame;
      frame.m_pageId=int(input->readLong(1));
      f << "page[id]=" << frame.m_pageId << ",";
      val=int(input->readULong(1));
      frame.m_type=val&3;
      f << "type=" << frame.m_type << ",";
      if (val&0x80) f << "crop[image],";
      val&=0x7c;
      // checkme what means type+0x10: multicolumn?
      f << "fl=" << std::hex << val << std::dec << ","; // 0: layout, 1: text, 2: graphic
      val=int(input->readULong(1));
      if (val==0x40)
        f << "crop[image],"; // checkme
      else if (val)
        f << "f0=" << val << ",";
      auto &style=frame.m_style;
      int bStyle=int(input->readLong(1));
      f << "border[style]=" << bStyle << ","; // 0: single, 1: double, 2: 2,1,1, 3: 1,1,2, 4: 1,1,2,1,1, 5: dashed, 6; dotted, 7: big dotted
      style.m_lineWidth=float(input->readLong(1));
      if (style.m_lineWidth>0) {
        f << "line[width]=" << style.m_lineWidth << ",";
        if (bStyle>=1 && bStyle<=4)
          style.m_lineWidth*=2;
        else if (bStyle==5)
          style.m_lineDashWidth= {10,10};
        else if (bStyle>=6 && bStyle<=7)
          style.m_lineDashWidth= {2,2};
      }
      val=int(input->readLong(1));
      if (val!=1) f << "line[pat]=" << val << ",";
      frame.m_page=int(input->readLong(2));
      f << "page=" << frame.m_page << ",";
      for (int j=0; j<5; ++j) { // f5=0|b, f6=0|12, f7=0|100
        val=int(input->readLong(2));
        if (!val)
          continue;
        if (j==1) {
          frame.m_numSubZones=val;
          f << "num[subZones]=" << val << ",";
        }
        else if (j==4)
          f << "gray[scale]=" << val << ",";
        else
          f << "f" << j+3 << "=" << val << ",";
      }
      for (int j=0; j<2; ++j) {
        val=int(input->readLong(2));
        if (val==-1)
          continue;
        f << "g" << j << "=" << val << ",";
      }
      val=int(input->readLong(4));
      if (val!=-1) f << "next=F"  << val << ",";
      if (m_state->m_idToFrameMap.count(i)==0)
        m_state->m_idToFrameMap[i]=frame;
      else {
        MWAW_DEBUG_MSG(("SpringBoardParser::parseData: find a dupplicated frame with id=%d\n", i));
        f << "###dup,";
      }
      break;
    }
    case 3: { // frame header/footer link
      f << "unkn=" << std::hex << input->readLong(2) << std::dec << ",";
      val=int(input->readULong(2));
      if (val!=1)
        f << "f0=" << val << ",";
      f << "fId=" << int(input->readLong(2)) << ",";
      val=int(input->readULong(2));
      if (val)
        f << "font[fl]=" << std::hex << val << std::dec << ",";
      f << "fSize=" << int(input->readLong(2)) << ",";
      for (int j=0; j<2; ++j) {
        long actPos=input->tell();
        int cLen=int(input->readULong(1));
        if (cLen>80) {
          MWAW_DEBUG_MSG(("SpringBoardParser::parseData: can not read a string\n"));
          f << "###cLen" << j << "=" << cLen << ",";
          cLen=0;
        }
        std::string text;
        for (int c=0; c<cLen; ++c) text+=char(input->readULong(1));
        if (!text.empty())
          f << (j==0 ? "top" : "bottom") << "[text]=" << text << ",";
        input->seek(actPos+80, librevenge::RVNG_SEEK_SET);
      }
      for (int j=0; j<3; ++j) {
        int pageId[2];
        for (auto &w : pageId) w=int(input->readLong(2));
        if (pageId[0]==-1 && pageId[1]==-1) continue;
        char const *wh[]= {"id", "prev[id]", "next[id]"};
        f << wh[j] << "=" << MWAWVec2i(pageId[0], pageId[1]) << ",";
      }
      break;
    }
    // 8: already done
    case 9: { // Text
      std::string text;
      for (int j=0; j<512; ++j) {
        char c=char(input->readULong(1));
        if (!c)
          break;
        text+=c;
      }
      f << text << ",";
      break;
    }
    default:
      break;
    }
    input->seek(pos+expectedSize[id], librevenge::RVNG_SEEK_SET);
    ascii().addPos(pos);
    ascii().addNote(f.str().c_str());
  }
  return true;
}

bool SpringBoardParser::parseParagraphStyle(MWAWEntry const &entry, MWAWVec2i const &indices)
{
  MWAWInputStreamPtr input = getInput();
  if (!input)
    return false;
  int const N=indices[1]-indices[0];
  if (!entry.valid() || !input->checkPosition(entry.end()) || N<0 || entry.length()<4*N) {
    MWAW_DEBUG_MSG(("SpringBoardParser::parseParagraphStyle: find invalid entry\n"));
    return false;
  }
  ascii().addPos(entry.end());
  ascii().addNote("_");

  input->seek(entry.begin(), librevenge::RVNG_SEEK_SET);
  long pos=input->tell();
  libmwaw::DebugStream f;
  f << "ParaStyle[" << N << "]:zones=[";
  std::map<int, long> ptrs;
  for (int i=indices[0]; i<indices[1]; ++i) {
    long ptr=long(input->readULong(4));
    f << std::hex << ptr << std::dec << ",";
    if (ptr>=entry.begin()+4*N && ptr+16<=entry.end())
      ptrs[i]=ptr;
    else {
      f << "###";
      MWAW_DEBUG_MSG(("SpringBoardParser::parseParagraphStyle: find bad ptr\n"));
    }
  }
  f << "],";
  ascii().addPos(pos);
  ascii().addNote(f.str().c_str());
  int val;
  for (auto const &it : ptrs) {
    f.str("");
    f << "ParaStyle-P" << it.first << ":";
    for (int i=0; i<4; ++i) {
      val=int(input->readULong(1));
      if (val==(i==3 ? 1 : 0)) continue;
      if (i==0)
        f << "align=" << val << ","; // 1: center, 2:right, 3: justify?
      else if (i==1)
        f << "line[bef]=" << val << "%,";
      else if (i==2)
        f << "line[after]=" << val << "%,";
      else
        f << "f0=" << val << ",";
    }

    for (int i=0; i<5; ++i) {
      val=int(input->readULong(2));
      int const expected[]= {1,0,0,0,0};
      if (val==expected[i])
        continue;
      if (i==1) { // more like after
        if (val&0x8000)
          f << "line[spacing]=" << (val&0x7fff) << ",";
        else
          f << "line[spacing]=" << (1+val) << "%";
      }
      else if (i==2)
        f << "first[indent]=" << val << ",";
      else if (i==3)
        f << "left[indent]=" << val << ",";
      else if (i==4)
        f << "right[indent]=" << val << ",";
    }
    int nTabs=int(input->readULong(2));
    if (nTabs) {
      f << "nTabs=" << nTabs << ",";
      if (pos+16+4*nTabs>entry.end()) {
        f << "###";
        MWAW_DEBUG_MSG(("SpringBoardParser::parseParagraphStyle: the number of tabs seems bad\n"));
        nTabs=0;
      }
      f << "tabs=[";
      for (int i=0; i<nTabs; ++i) {
        char c=char(input->readLong(1));
        input->seek(1, librevenge::RVNG_SEEK_CUR);
        f << input->readLong(2) << ":" << c << ",";
      }
      f << "],";
    }
    ascii().addPos(it.second);
    ascii().addNote(f.str().c_str());
  }
  return true;
}

bool SpringBoardParser::parseLastZone(int id, MWAWEntry const &entry)
{
  MWAWInputStreamPtr input = getInput();
  if (!input)
    return false;
  if (id<0 || id>=3 || !entry.valid() || !input->checkPosition(entry.end())) {
    MWAW_DEBUG_MSG(("SpringBoardParser::parseLastZone: find invalid entry for id=%d\n", id));
    return false;
  }
  ascii().addPos(entry.end());
  ascii().addNote("_");

  libmwaw::DebugStream f, f2;
  char const *wh[]= {"Shape", "TxtPos", "Frame"};
  std::string const what(wh[id]);
  f << "Entries(" << what << "):";
  input->seek(entry.begin(), librevenge::RVNG_SEEK_SET);
  switch (id) {
  case 0: {
    if ((entry.length()%8)!=0) {
      MWAW_DEBUG_MSG(("SpringBoardParser::parseLastZone: unexpected size\n"));
      f << "###";
      break;
    }
    int N=int(entry.length()/8);
    std::map<int,long> idToPtrs;
    for (int i=0; i<N; ++i) {
      long pos=input->tell();
      f2.str("");
      f2 << what << "-" << i << ":";
      long beg=long(input->readULong(4));
      f2 << "pos=" << std::hex << beg << std::dec << ",";
      if (beg<0x80 || !input->checkPosition(beg+4)) {
        MWAW_DEBUG_MSG(("SpringBoardParser::parseLastZone: a shape position seems bad\n"));
        f2 << "###";
      }
      else
        idToPtrs[i]=beg;
      f2 << "page=" << input->readLong(2) << ",";
      f2 << "id=" << input->readLong(2) << ",";
      ascii().addPos(pos);
      ascii().addNote(f2.str().c_str());
    }
    for (auto const &it : idToPtrs)
      parseShape(it.second, it.first);
    break;
  }
  case 1: { // checkme maybe link
    if ((entry.length()%6)!=0) {
      MWAW_DEBUG_MSG(("SpringBoardParser::parseLastZone: unexpected size\n"));
      f << "###";
      break;
    }
    int N=int(entry.length()/6);
    f << "unkn=[";
    for (int i=0; i<N; ++i) {
      int val=int(input->readLong(4));
      f << "last[char]=" << input->readLong(2);
      if (val!=-1)
        f << ":" << val;
      f << ",";
    }
    f << "],";
    break;
  }
  case 2: {
    f << "list,";
    if (entry.length()<2) {
      MWAW_DEBUG_MSG(("SpringBoardParser::parseLastZone: unexpected size\n"));
      f << "###";
      break;
    }
    int N=int(input->readLong(2));
    f << "N=" << N << ",";
    if (2+8*N>entry.length()) {
      MWAW_DEBUG_MSG(("SpringBoardParser::parseLastZone: unexpected N\n"));
      f << "###";
      break;
    }
    std::map<int,long> positions;
    for (int i=0; i<N; ++i) {
      long pos=input->tell();
      f2.str("");
      f2 << what << "-" << i << "[list]:";
      int pg=int(input->readLong(2));
      f2 << "page=" << pg << ",";
      int val=int(input->readLong(2));
      if (val!=pg) f2 << "page2=" << val << ",";
      long beg=long(input->readULong(4));
      f2 << "pos=" << std::hex << beg << std::dec << ",";
      if (beg<0x80 || !input->checkPosition(beg)) {
        MWAW_DEBUG_MSG(("SpringBoardParser::parseLastZone: a position seems bad\n"));
        f << "###";
      }
      else
        positions[pg]=beg;
      ascii().addPos(pos);
      ascii().addNote(f2.str().c_str());
    }
    for (auto const &it : positions)
      parseFrames(it.first, it.second);
    break;
  }
  default:
    break;
  }
  ascii().addPos(entry.begin());
  ascii().addNote(f.str().c_str());
  return true;
}

bool SpringBoardParser::parseShape(long begPos, int frameId)
{
  MWAWInputStreamPtr input = getInput();
  if (!input)
    return false;
  if (begPos<0x80 || !input->checkPosition(begPos+4)) {
    MWAW_DEBUG_MSG(("SpringBoardParser::parseShape: the begin position seems bad\n"));
    return false;
  }
  input->seek(begPos, librevenge::RVNG_SEEK_SET);
  auto const fIt=m_state->m_idToFrameMap.find(frameId);
  int type=fIt!=m_state->m_idToFrameMap.end() ? fIt->second.m_type : -1;

  libmwaw::DebugStream f;
  f << "Shape-S" << frameId << "[type=" << type << "]:";
  input->seek(begPos, librevenge::RVNG_SEEK_SET);

  switch (type) {
  case 0:
  case 1: {
    int N=int(input->readULong(4));
    f << "N=" << N << ",";
    if (N<0 || N+1>(input->size()-begPos)/4 || !input->checkPosition(4+4*N)) {
      MWAW_DEBUG_MSG(("SpringBoardParser::parseShape: can not read the number of shapes\n"));
      f << "###";
      N=0;
    }
    std::vector<long> listPtrs;
    for (int i=0; i<N; ++i) listPtrs.push_back(long(input->readULong(4)));
    ascii().addPos(begPos);
    ascii().addNote(f.str().c_str());

    for (long ptr : listPtrs) {
      if (ptr<0x80 || !input->checkPosition(ptr+24)) {
        MWAW_DEBUG_MSG(("SpringBoardParser::parseShape: a text pointer seems bad\n"));
        continue;
      }
      f.str("");
      f << "Shape-S" << frameId << ":text,";
      input->seek(ptr, librevenge::RVNG_SEEK_SET);
      auto nChar=input->readULong(4);
      if (nChar&0x80000000)
        f << "fl8,";
      nChar&=0x7fffffff;
      f << "num[char]=" << nChar << ",";
      int val;
      for (int i=0; i<2; ++i) {
        val=int(input->readLong(2));
        if (!val) continue;
        if (i==0)
          f << "first[char]=" << val << ",";
        else
          f << "y[pos]=" << val << ",";
      }
      val=int(input->readLong(4));
      if (val!=ptr) f << "ptr=" << std::hex << val << std::dec << ",";
      int numCharStyle=int(input->readLong(4));
      f << "num[style]=" << numCharStyle << ",";
      f << "para[style]=" << input->readLong(4) << ",";
      val=int(input->readLong(4));
      if (val!=-1)
        f << "text[id]=" << val << ",";
      if (ptr+24+10*numCharStyle<ptr+24 || (input->size()-24-ptr)/10<numCharStyle ||
          !input->checkPosition(ptr+24+10*numCharStyle)) {
        f << "###";
        MWAW_DEBUG_MSG(("SpringBoardParser::parseShape: the number of style seems bad\n"));
        numCharStyle=0;
      }
      ascii().addPos(ptr);
      ascii().addNote(f.str().c_str());
      int cPos=0;
      for (int i=0; i<numCharStyle; ++i) {
        long pos=input->tell();
        f.str("");
        f << "Shape-S" << frameId << "[" << i << "]:";
        f << "cPos=" << cPos << ",";
        cPos+=int(input->readLong(2));
        f << "fId=" << input->readULong(2) << ",";
        val=int(input->readLong(2));
        if (val) f << "fl=" << val << ",";
        f << "fSz=" << input->readULong(1) << ",";
        val=int(input->readLong(2));
        if (val) f << "char[spacing]=" << val << ",";
        input->seek(pos+10, librevenge::RVNG_SEEK_SET);
        ascii().addPos(pos);
        ascii().addNote(f.str().c_str());
      }
    }
    return true;
  }
  case 2: {
    int val=int(input->readULong(4));
    if (val)
      f << "f0=" << val << ",";
    long len=long(input->readULong(4));
    f << "len=" << len << ",";
    if ((len && len<10) || begPos+12+len<begPos+12 || !input->checkPosition(begPos+12+len)) {
      MWAW_DEBUG_MSG(("SpringBoardParser::parseShape: can not read the picture length\n"));
      f << "###";
      break;
    }
    val=int(input->readULong(4));
    if (val)
      f << "f1=" << val << ",";
    if (!len)
      break;
#ifdef DEBUG_WITH_FILES
    std::shared_ptr<MWAWPict> pict(MWAWPictData::get(input, int(len)));
    MWAWEmbeddedObject image;
    if (pict && pict->getBinary(image) && !image.m_dataList.empty()) {
      static int volatile pictName = 0;
      libmwaw::DebugStream f2;
      f2 << "PICT-" << ++pictName << ".pct";
      libmwaw::Debug::dumpFile(image.m_dataList[0], f2.str().c_str());
      ascii().skipZone(begPos+12, begPos+12+len-1);
    }
    else
#endif
      MWAW_DEBUG_MSG(("SpringBoardParser::parseShape: sorry, can not retrieve a picture\n"));
    break;
  }
  default:
    MWAW_DEBUG_MSG(("SpringBoardParser::parseShape: find unknown type of shapes\n"));
    f << "###";
    break;
  }
  ascii().addPos(begPos);
  ascii().addNote(f.str().c_str());
  return true;
}

bool SpringBoardParser::parseFrames(int page, long begPos)
{
  MWAWInputStreamPtr input = getInput();
  if (!input)
    return false;
  if (begPos<0x80 || !input->checkPosition(begPos+14)) {
    MWAW_DEBUG_MSG(("SpringBoardParser::parseFrames: the begin position seems bad\n"));
    return false;
  }
  libmwaw::DebugStream f;
  f << "Entries(Frame):page=" << page << ",";
  input->seek(begPos, librevenge::RVNG_SEEK_SET);
  int val;
  for (int i=0; i<2; ++i) { // f0=0|-4|10
    val=int(input->readLong(2));
    if (val!=i)
      f << "f" << i << "=" << val << ",";
  }
  int dim[4];
  for (auto &d : dim) d=int(input->readLong(2));
  f << "box=" << MWAWBox2i(MWAWVec2i(dim[1], dim[0]),MWAWVec2i(dim[3], dim[2])) << ",";
  int N=int(input->readULong(2));
  f << "N=" << N << ",";
  if (!input->checkPosition(14+N*28)) {
    MWAW_DEBUG_MSG(("SpringBoardParser::parseFrames: can not read N\n"));
    f << "###";
    ascii().addPos(begPos);
    ascii().addNote(f.str().c_str());
    return false;
  }
  ascii().addPos(begPos);
  ascii().addNote(f.str().c_str());

  for (int i=0; i<N; ++i) {
    // checkme: seems related to Zone2A
    long pos=input->tell();
    f.str("");
    int id=int(input->readLong(1));
    f << "Frame-" << page << "x" << id << ":";
    val=int(input->readLong(1));
    switch (val) {
    case 0:
      f << "layout,";
      break;
    case 1:
      f << "text,";
      break;
    case 2:
      f << "graphics,";
      break;
    default:
      MWAW_DEBUG_MSG(("SpringBoardParser::parseFrames: find unknown frame type\n"));
      f << "###type=" << val << ",";
      break;
    }
    for (auto &d : dim) d=int(input->readLong(2));
    f << "box=" << MWAWBox2i(MWAWVec2i(dim[1], dim[0]),MWAWVec2i(dim[3], dim[2])) << ",";
    for (auto &d : dim) d=int(input->readLong(2));
    f << "box[internal]=" << MWAWBox2i(MWAWVec2i(dim[1], dim[0]),MWAWVec2i(dim[3], dim[2])) << ",";
    for (int j=0; j<4; ++j) {
      val=int(input->readLong(2));
      if (val==0) continue;
      f << "f" << j << "=" << val << ",";
    }
    int n1=int(input->readULong(2));
    if (n1!=1)
      f << "N[col]=" << n1 << ",";
    if (!input->checkPosition(pos+28+n1*10)) {
      MWAW_DEBUG_MSG(("SpringBoardParser::parseFrames: find unexpected subzone N\n"));
      f << "###";
      ascii().addPos(pos);
      ascii().addNote(f.str().c_str());
      return false;
    }
    ascii().addPos(pos);
    ascii().addNote(f.str().c_str());

    for (int col=0; col<n1; ++col) {
      pos=input->tell();
      f.str("");
      f << "Frame-F" << i << "[" << col << "]:";
      val=int(input->readULong(1));
      if (val)
        f << "f0=" << val << ",";
      val=int(input->readULong(1));
      if (val!=1)
        f << "col=" << val << ",";
      for (auto &d : dim) d=int(input->readLong(2));
      f << "box=" << MWAWBox2i(MWAWVec2i(dim[1], dim[0]),MWAWVec2i(dim[3], dim[2])) << ",";
      ascii().addPos(pos);
      ascii().addNote(f.str().c_str());
    }
  }
  return true;
}

////////////////////////////////////////////////////////////
// send shapes
////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////
// read shapes
////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////
// read the header
////////////////////////////////////////////////////////////
bool SpringBoardParser::checkHeader(MWAWHeader *header, bool strict)
{
  *m_state = SpringBoardParserInternal::State();
  MWAWInputStreamPtr input = getInput();
  if (!input || !input->hasDataFork() || !input->checkPosition(128))
    return false;

  libmwaw::DebugStream f;
  f << "FileHeader:";
  input->seek(0, librevenge::RVNG_SEEK_SET);
  if (input->readULong(2)!=0xcd90)
    return false;
  for (int i=0; i<2; ++i) {
    int val=int(input->readLong(1));
    if (val==i) continue;
    if (i==0)
      f << "rev" << val << ","; // look like a counter (modulo 10)
    else
      f << "f" << i << "=" << val << ",";
  }
  if ((input->readULong(2)&0xff00)!=0) return false;

  if (strict) {
  }

  ascii().addPos(0);
  ascii().addNote(f.str().c_str());
  if (header)
    header->reset(MWAWDocument::MWAW_T_SPRINGBOARD_PUBLISHER, 2, MWAWDocument::MWAW_K_DRAW);
  input->seek(4, librevenge::RVNG_SEEK_SET);

  return true;
}

// vim: set filetype=cpp tabstop=2 shiftwidth=2 cindent autoindent smartindent noexpandtab:
